@@ -4,7 +4,7 @@ use crate::{
     attributes::FromPyWithAttribute,
     method::{FnArg, FnSpec},
     pyfunction::Argument,
-    utils::unwrap_ty_group,
+    utils::{remove_lifetime, replace_self, unwrap_ty_group},
 };
 use proc_macro2::{Span, TokenStream};
 use quote::{quote, quote_spanned};
@@ -78,6 +78,7 @@ pub fn impl_arg_params(
     };
 
     let mut positional_parameter_names = Vec::new();
+    let mut positional_only_parameters = 0usize;
     let mut required_positional_parameters = 0usize;
     let mut keyword_only_parameters = Vec::new();
 
@@ -86,12 +87,13 @@ pub fn impl_arg_params(
             continue;
         }
         let name = arg.name.unraw().to_string();
+        let posonly = spec.is_pos_only(arg.name);
         let kwonly = spec.is_kw_only(arg.name);
         let required = !(arg.optional.is_some() || spec.default_value(arg.name).is_some());
 
         if kwonly {
             keyword_only_parameters.push(quote! {
-                pyo3::derive_utils::KeywordOnlyParameterDescription {
+                ::pyo3::derive_utils::KeywordOnlyParameterDescription {
                     name: #name,
                     required: #required,
                 }
@@ -100,6 +102,9 @@ pub fn impl_arg_params(
             if required {
                 required_positional_parameters += 1;
             }
+            if posonly {
+                positional_only_parameters += 1;
+            }
             positional_parameter_names.push(name);
         }
     }
@@ -123,9 +128,9 @@ pub fn impl_arg_params(
     let (accept_args, accept_kwargs) = accept_args_kwargs(&spec.attrs);
 
     let cls_name = if let Some(cls) = self_ {
-        quote! { Some(<#cls as pyo3::type_object::PyTypeInfo>::NAME) }
+        quote! { ::std::option::Option::Some(<#cls as ::pyo3::type_object::PyTypeInfo>::NAME) }
     } else {
-        quote! { None }
+        quote! { ::std::option::Option::None }
     };
     let python_name = &spec.python_name;
 
@@ -134,8 +139,9 @@ pub fn impl_arg_params(
         // keyword names of the keyword args in _kwargs
         (
             // need copied() for &&PyAny -> &PyAny
-            quote! { _args.iter().copied() },
+            quote! { ::std::iter::Iterator::copied(_args.iter()) },
             quote! { _kwnames.map(|kwnames| {
+                use ::std::iter::Iterator;
                 kwnames.as_slice().iter().copied().zip(_kwargs.iter().copied())
             }) },
         )
@@ -149,19 +155,18 @@ pub fn impl_arg_params(
 
     // create array of arguments, and then parse
     Ok(quote! {{
-            const DESCRIPTION: pyo3::derive_utils::FunctionDescription = pyo3::derive_utils::FunctionDescription {
+            const DESCRIPTION: ::pyo3::derive_utils::FunctionDescription = ::pyo3::derive_utils::FunctionDescription {
                 cls_name: #cls_name,
                 func_name: stringify!(#python_name),
                 positional_parameter_names: &[#(#positional_parameter_names),*],
-                // TODO: https://github.com/PyO3/pyo3/issues/1439 - support specifying these
-                positional_only_parameters: 0,
+                positional_only_parameters: #positional_only_parameters,
                 required_positional_parameters: #required_positional_parameters,
                 keyword_only_parameters: &[#(#keyword_only_parameters),*],
                 accept_varargs: #accept_args,
                 accept_varkeywords: #accept_kwargs,
             };
 
-            let mut #args_array = [None; #num_params];
+            let mut #args_array = [::std::option::Option::None; #num_params];
             let (_args, _kwargs) = DESCRIPTION.extract_arguments(
                 #py,
                 #args_to_extract,
@@ -201,7 +206,7 @@ fn impl_arg_param(
     let ty = arg.ty;
     let name = arg.name;
     let transform_error = quote! {
-        |e| pyo3::derive_utils::argument_extraction_error(#py, stringify!(#name), e)
+        |e| ::pyo3::derive_utils::argument_extraction_error(#py, stringify!(#name), e)
     };
 
     if is_args(&spec.attrs, name) {
@@ -235,12 +240,22 @@ fn impl_arg_param(
 
     let arg_value_or_default = match (spec.default_value(name), arg.optional.is_some()) {
         (Some(default), true) if default.to_string() != "None" => {
-            quote_arg_span! { #arg_value.map_or_else(|| Ok(Some(#default)), |_obj| #extract)? }
+            quote_arg_span! {
+                #arg_value.map_or_else(|| ::std::result::Result::Ok(::std::option::Option::Some(#default)),
+                                       |_obj| #extract)?
+            }
         }
         (Some(default), _) => {
-            quote_arg_span! { #arg_value.map_or_else(|| Ok(#default), |_obj| #extract)? }
+            quote_arg_span! {
+                #arg_value.map_or_else(|| ::std::result::Result::Ok(#default), |_obj| #extract)?
+            }
+        }
+        (None, true) => {
+            quote_arg_span! {
+                #arg_value.map_or(::std::result::Result::Ok(::std::option::Option::None),
+                                  |_obj| #extract)?
+            }
         }
-        (None, true) => quote_arg_span! { #arg_value.map_or(Ok(None), |_obj| #extract)? },
         (None, false) => {
             quote_arg_span! {
                 {
@@ -252,11 +267,15 @@ fn impl_arg_param(
     };
 
     return if let syn::Type::Reference(tref) = unwrap_ty_group(arg.optional.unwrap_or(ty)) {
-        let (tref, mut_) = preprocess_tref(tref, self_);
+        let mut tref = remove_lifetime(tref);
+        if let Some(cls) = self_ {
+            replace_self(&mut tref.elem, cls);
+        }
+        let mut_ = tref.mutability;
         let (target_ty, borrow_tmp) = if arg.optional.is_some() {
             // Get Option<&T> from Option<PyRef<T>>
             (
-                quote_arg_span! { Option<<#tref as pyo3::derive_utils::ExtractExt<'_>>::Target> },
+                quote_arg_span! { ::std::option::Option<<#tref as ::pyo3::derive_utils::ExtractExt<'_>>::Target> },
                 if mut_.is_some() {
                     quote_arg_span! { _tmp.as_deref_mut() }
                 } else {
@@ -266,7 +285,7 @@ fn impl_arg_param(
         } else {
             // Get &T from PyRef<T>
             (
-                quote_arg_span! { <#tref as pyo3::derive_utils::ExtractExt<'_>>::Target },
+                quote_arg_span! { <#tref as ::pyo3::derive_utils::ExtractExt<'_>>::Target },
                 quote_arg_span! { &#mut_ *_tmp },
             )
         };
@@ -280,33 +299,4 @@ fn impl_arg_param(
             let #arg_name = #arg_value_or_default;
         })
     };
-
-    /// Replace `Self`, remove lifetime and get mutability from the type
-    fn preprocess_tref(
-        tref: &syn::TypeReference,
-        self_: Option<&syn::Type>,
-    ) -> (syn::TypeReference, Option<syn::token::Mut>) {
-        let mut tref = tref.to_owned();
-        if let Some(syn::Type::Path(tpath)) = self_ {
-            replace_self(&mut tref, &tpath.path);
-        }
-        tref.lifetime = None;
-        let mut_ = tref.mutability;
-        (tref, mut_)
-    }
-
-    /// Replace `Self` with the exact type name since it is used out of the impl block
-    fn replace_self(tref: &mut syn::TypeReference, self_path: &syn::Path) {
-        match &mut *tref.elem {
-            syn::Type::Reference(tref_inner) => replace_self(tref_inner, self_path),
-            syn::Type::Path(tpath) => {
-                if let Some(ident) = tpath.path.get_ident() {
-                    if ident == "Self" {
-                        tpath.path = self_path.to_owned();
-                    }
-                }
-            }
-            _ => {}
-        }
-    }
 }
